local require = require

require("eclass")
local ffi = require("ffi")
local system = require("common.system")
local md5 = require("md5")
local pystring = require("pystring")
local cjson = require("cjson")
local module = require("app.module")
local api = require("haWork.api")
local plain = require("haWork.plain")
local liveTrace = require("liveTrace.liveTrace")

local format = string.format
local print = print
local tonumber = tonumber
local match = string.match
local io_open = io.open
local unpack = unpack
local split = pystring.split
local sumhexa = md5.sumhexa
local jdecode = cjson.decode
local jencode = cjson.encode
local liveStart = liveTrace.start
local liveStop = liveTrace.stop
local os_remove = os.remove
local haApi = api.api
local haPlain = plain.plain
local collectgarbage = collectgarbage

collectgarbage("setpause", 150)
collectgarbage("setstepmul", 300)

local c_api = ffi.load('libbpfpre.so')
ffi.cdef[[
    int bump_memlock_rlimit1(void);
]]

local class = class
local Cmain = class("test")

local c_bump_memlock_rlimit1 = c_api.bump_memlock_rlimit1
local counter = 0
local localModule = module.new("plugin")
local startArgs

local function index(tReq)
    counter = counter + 1
    return {body = format("beaver kata %d say hello.", counter)}
end

local function elfsym(tReq)    -- body
    local range = tReq.headers["range"]
    if not range then
        system.dumps(tReq.headers)
        return {code = 400, body = "Range is nil."}
    end

    local start, stop = match(range, "bytes=(%d+)-(%d+)")
    start, stop = tonumber(start), tonumber(stop)
    if start and stop and stop > start then
        local size = stop - start + 1
        local f = io_open("/tmp/elfsym", "rb")
        if f then
            f:seek("set", start)
            local s = f:read(size)
            local full = f:seek("end")
            f:close()
            return {code = 206, body = s, headers = {["Content-Range"] = format("bytes %d-%d/%d", start, stop, full)}}
        else
            return {code = 400, body = format("open file /tmp/elfsym failed.")}
        end
    else
        return {code = 400, body = format("Range %s is not valid.", range)}
    end
end

local function deElfsym(tReq)
    local ok, msg = os_remove("/tmp/elfsym")
    if not ok then
        return {code = 503, body = format("remove file /tmp/elfsym failed, report: %s.", msg)}
    else
        return {body = "remove file /tmp/elfsym ok."}
    end
end

local function saveFile(tReq)
    local path = tReq.headers["file-path"]
    local wpath
    local _, suffix = unpack(split(path, ".", 1))
    if suffix == "lua" then
        wpath = format("%s/%s", startArgs.lua_path, path)
    elseif suffix == "so" then
        wpath = format("%s/%s", startArgs.so_path, path)
    else
        return {code = 400, body = "file type not support."}
    end

    local body = tReq.body
    if tReq.headers["content-md5"] ~= sumhexa(body) then
        return {code = 400, body = "md5 not match."}
    end

    local f = io_open(wpath, "wb")
    if not f then
        return {code = 400, body = format("open file %s failed.", wpath)}
    end
    f:write(body)
    f:close()

    return {body = format("save file:%s ok.", wpath)}
end

local function upload(tReq)
    return saveFile(tReq)
end

local function moduleMan(tReq)
    local args = jdecode(tReq.body)
    if not args then
        return {code = 400, body = "json decode failed."}
    end

    local ret, err = localModule:call(args)
    if not ret then
        return {code = 400, body = err}
    end

    return {body = "module manage ok."}
end

local function liveStartPost(tReq)
    local body = tReq.body

    local res = jdecode(body)
    if not res then
        return {body = "body is not json."}
    end

    return liveStart(res.module, res.args)
end

local function liveStopPost(tReq)
    local body = tReq.body

    local res = jdecode(body)
    if not res then
        return {body = "body is not json."}
    end

    return liveStop(res.module)
end

function Cmain:_init_(inst, conf)
    c_bump_memlock_rlimit1()
    startArgs = conf.args
    inst:get("/", index)
    inst:get("/elfsym", elfsym)
    inst:delete("/elfsym", deElfsym)
    inst:put("/upload", upload)
    inst:post("/module", moduleMan)
    inst:post("/api", haApi)
    inst:post("/plain", haPlain)
    inst:post("/liveStart", liveStartPost)
    inst:post("/liveStop", liveStopPost)
end

return Cmain
