local require = require

require("eclass")
local system = require("common.system")

local class = class
local CModule = class("module")

local package = package
local format = string.format

function CModule:_init_(path)
    self._modules = {}
    self._path = path
end

function CModule:_del_()
end

function CModule:add(name, args)
    local modules = self._modules
    if modules[name] then
        return nil, format("module %s already exists", name)
    end
    local m = require(format("%s.%s", self._path, name))
    if not m then
        return nil, "module not found"
    end
    modules[name] = m.new(args)
    return m
end

function CModule:del(name, args)
    local modules = self._modules
    local m = modules[name]
    if not m then
        return nil, format("module %s not found", name)
    end
    if m:status() then
        return nil, format("module %s is running", name)
    end

    modules[name] = nil
    local path = format("%s.%s", self._path, name)
    package.loaded[path] = nil
    return m
end

function CModule:start(name, args)
    local m = self._modules[name]
    if not m then
        return nil, format("module %s not found", name)
    end
    return m:start(args)
end

function CModule:stop(name, args)
    local m = self._modules[name]
    if not m then
        return nil, format("module %s not found", name)
    end
    return m:stop(args)
end

local cbs = {
    ["start"] = CModule.start,
    ["stop"] = CModule.stop,
    ["add"] = CModule.add,
    ["del"] = CModule.del,
}

function CModule:call(args)
    local func = args["func"]
    if cbs[func] then
        return cbs[func](self, args.args)
    else
        return nil, format("func %s not found", func)
    end
end

return CModule
