---
--- Generated by EmmyLua(https://github.com/EmmyLua)
--- Created by liaozhaoyan.
--- DateTime: 2024/1/17 2:34 PM
---

local require = require
require("eclass")
local ffi = require("ffi")
local ptime = require("posix.time")
local pwait = require("posix.sys.wait")
local signal = require("posix.signal")
local unistd = require("posix.unistd")
local struct = require("struct")

local class = class
local CbaseFork = class("baseFork")

local assert = assert
local error = error
local type = type
local print = print
local upper = string.upper
local concat = table.concat
local sub = string.sub
local exit = os.exit
local unistd_pipe = unistd.pipe
local unistd_fork = unistd.fork
local kill = signal.kill
local write = unistd.write
local read = unistd.read
local close = unistd.close
local nanosleep = ptime.nanosleep
local pwait_wait = pwait.wait
local struct_pack = struct.pack
local struct_unpack = struct.unpack

ffi.cdef[[
int prctl(int option, unsigned long arg2, unsigned long arg3, unsigned long arg4, unsigned long arg5);
]]

local function setCommName(name)
    assert(type(name) == "string", "comm argument must be string")
    if #name > 16 then
        print("warnning: comm name too long.", name)
    end
    local res = ffi.C.prctl(15, ffi.cast("unsigned long", name), 0, 0, 0)
    if res ~= 0 then
        error("Failed to set process name")
    end
end

function CbaseFork:_init_(name)
    local sr, cw = unistd_pipe()  -- for server read
    local cr, sw = unistd_pipe()  -- for server write
    local childPid = unistd_fork()
    if childPid == 0 then
        close(sr)
        close(sw)
        if name then
            setCommName(name)
        end
        self:_main(cr, cw)
        close(cr)
        close(cw)
        print("child exit.")
        exit(0)
    else
        close(cr)
        close(cw)
        self._childPid = childPid
        self._sr = sr
        self._sw = sw
    end
end

local function yield(ns)
    ns = ns or 0
    nanosleep({tv_sec=0, tv_nsec=ns})
end

function CbaseFork:_del_()  -- recycle resource
    if self._sr then
        close(self._sr)
    end
    if self._sw then
        close(self._sw)
    end
    if self._childPid then
        local childPid, stat = pwait_wait(self._childPid, pwait.WNOHANG)
        if stat == "running" then  -- runing? just wait 1ms to check.
            yield(1e6)
            childPid, stat = pwait_wait(childPid, pwait.WNOHANG)
            if stat == "running" then  -- running? killed.
                kill(childPid, signal.SIGTERM)
                yield(1000)
                childPid, stat = pwait_wait(self._childPid, pwait.WNOHANG)
                if stat == "running" then  -- still running killed.
                    kill(childPid, signal.SIGKILL)
                end
            end
        end
    end
end

function CbaseFork:pipe()
    return self._r, self._w
end

function CbaseFork:pipeWrite(fd, stream)
    local sent, err, errno
    local len = #stream
    local buff = concat({struct_pack("<i", len), stream})

    len = len + 4
    while len > 0 do
        sent, err, errno = write(fd, buff)
        if sent then
            len = len - sent
            buff = sub(buff, sent + 1)
        else
            break
        end
    end

    if errno then
        return nil, err, errno
    end
    return 0
end

function CbaseFork:pipeRead(fd)
    local cells = {}
    local c = 1
    local s
    local res, err, errno

    res, err, errno = read(fd, 4)
    if #res == 4 then
        local len = struct_unpack("<i", res)
        while len > 0 do
            s = read(fd, len)
            if s then
                len = len - #s
                cells[c] = s
                c = c + 1
            else
                break
            end
        end
    end

    if errno then
        return nil, err, errno
    end
    if #cells > 0 then
        return concat(cells)
    else
        return nil, "child may exist.", -1
    end
end

function CbaseFork:pwrite(stream)
    return self:pipeWrite(self._sw, stream)
end

function CbaseFork:pread()
    return self:pipeRead(self._sr)
end

function CbaseFork:_main(r, w)
    while true do
        local s = self:pipeRead(r)
        if s == "stop" or s == nil then
            print("host stop.")
            break
        end

        self:pipeWrite(w, upper(s))
    end
end

return CbaseFork
