local require = require
local pystring = require("pystring")
local ffi = require("ffi")

ffi.cdef[[
    // strtoull 函数声明
    unsigned long long int strtoull(const char *str, char **endptr, int base);
]]

local split = pystring.split
local pyfind = pystring.find
local format = string.format
local ioLines = io.lines
local tonumber = tonumber
local strtoull = ffi.C.strtoull

local mt = {}

function mt.loadSyms()
    local bias
    local c = 1
    local res = {}
    local lastSym, lastAddr

    for line in ioLines("/proc/kallsyms") do
        if not bias then
            if pyfind(line, " T startup_") > 0 then
                local segs = split(line)
                local start = strtoull(segs[1], nil, 16)
                bias = start
                lastSym = segs[3]
                lastAddr = start - bias
            end
        else
            local segs = split(line)
            if segs[2] == "t" or segs[2] == "T" then
                local addr = strtoull(segs[1], nil, 16) - bias
                if addr > lastAddr then
                    res[c] = {tonumber(lastAddr), lastSym}
                    c = c + 1
                    if #segs > 3 then
                        lastSym = format("%s%s", segs[3], segs[4])
                    else
                        lastSym = segs[3]
                    end
                    lastAddr = addr
                end
            end
        end
    end
    res[c] = {tonumber(lastAddr), lastSym}
    res[c + 1] = {tonumber(lastAddr) + 1000, "-"}
    return res, bias
end

return mt
