local require = require
local ffi = require("ffi")

local open = io.open
local format = string.format
local print = print
local c_cast = ffi.cast
local c_string = ffi.string
local tonumber = tonumber

local cffi = ffi.load("haSyms")

ffi.cdef[[
    typedef int (*elf_call_t)(const char *name, unsigned long start, unsigned long stop);
    int ha_elf_syms(const char* path, elf_call_t call);
]]
local ha_elf_syms = cffi.ha_elf_syms

local mt = {}
mt["$lib"] = cffi   -- to anti gc dlclose cffi

function mt.elf_syms(path, f, limit)
    if not f then
        return -1
    end
    limit = limit or 25 * 1024 * 1024

    local total = -1  -- strip first -1
    local function elf_call_cb(name, start, stop)
        if total > 0 then
            if not f:write("\n") then
                return -1
            end
        end
        local s = format("%s %d %d", c_string(name), tonumber(start), tonumber(stop))
        local size = #s + 1

        if total + size > limit then
            return -1
        end
        total = total + size
        if not f:write(s) then
            return -1
        end
        return 0
    end
    local _cb = c_cast("elf_call_t", elf_call_cb)
    local ret = ha_elf_syms(path, _cb)
    _cb:free()
    if ret ~= 0 then
        return -1
    end
    return total
end

return mt
