local require = require
require("eclass")

local nsenter = require("nsenter")
local elfmap = require("elfmap")
local pystring = require("pystring")
local CbaseFork = require("haWork.baseFork")
local haSymOut = require("haWork.haSymOut")
local gc = require("haWork.gc")

local pcall = pcall
local print = print
local class = class
local unpack = unpack
local tostring = tostring
local io_open = io.open
local io_lines = io.lines
local format = string.format
local concat = table.concat
local split = pystring.split
local gfree = gc.free
local elf_syms = haSymOut.elf_syms

local CplainSyms = class("ChaElfSyms", CbaseFork)

function CplainSyms:_init_(name)
    CbaseFork._init_(self, name)
end

local function plainMap(elf, f)
    return elf_syms(elf, f)
end

function CplainSyms:_main(r, w)
    local ns = nsenter.new()

    while true do
        local s = self:pipeRead(r)
        if #s == 0 then
            break
        end
        local rStr = "-1"
        local pid, elf = unpack(split(s, ",", 1))
        if pid and elf then
            local f = io_open("/tmp/elfsym", "w")
            if f then
                local res, err = pcall(ns.enter, ns, pid, "mnt")
                if res then
                    local ok, nRes = pcall(plainMap, elf, f)
                    if ok then
                        rStr = tostring(nRes)
                    else
                        print("newMap failed", pid, elf, err)
                    end
                    ns:exit()
                else
                    print("nsenter failed", err, pid)
                end
                f:close()
            end
        else
            print("elfdump exit.")
            break
        end
        self:pipeWrite(w, rStr)
        gfree()
    end
    print("fork exit.")
end

function CplainSyms:getElfs(pid, elf)
    if elf == "/proc/kallsyms" then
        local f = io_open("/tmp/elfsym", "w")
        if f then
            local size = 0
            for line in io_lines("/proc/kallsyms") do
                if size > 0 then
                    f:write("\n")
                    size = size + 1
                end
                f:write(line)
                size = size + #line
            end
            f:close()
            return tostring(size)
        else
            return "-1"
        end
    end
    local send = format("%s,%s", pid, elf)
    self:pwrite(send)
    local s = self:pread()
    return s
end

return CplainSyms