local require = require
require("eclass")
local ffi = require("ffi")
local system = require("common.system")

local pairs = pairs
local open = io.open
local c_new = ffi.new
local reverseTable = system.reverseTable
local tostring = tostring
local tonumber = tonumber
local class = class

local CbackTrace = class("backTrace")

function CbackTrace:_init_(cells, c_stacks, args)
    self._prefix = args.storePrefix
    self._cells = cells
    self._c_stacks = c_stacks
end

function CbackTrace:stop()
end

local function pidExist(pid)
    local path = "/proc/" .. tostring(pid) .. "/comm"
    local f, _ = open(path, 'r')
    if f then
        local comm = f:read()
        f:close()
        return comm
    end
    return false
end

function CbackTrace:back(stacks)
    local c_stacks = self._c_stacks
    local cells = {}
    local cnt = 1
    local overFlow = false

    local pidExistArr = {[0] = "idle"}
    for key, v in pairs(stacks) do
        local cell

        local pid = tonumber(key.pid)   --- process pid
        local res = pidExistArr[pid]
        if res then
            cell = {pid, res}
        elseif res == false then
            goto continue
        else
            res = pidExist(pid)
            pidExistArr[pid] = res
            if res then
                cell = {pid, res}
            else
                goto continue
            end
        end    -- done

        local k_id = key.k_stack   -- process k_stack
        local k_arr = {}
        if k_id > 0 then
            local k_stacks = c_new("struct stack_addrs")
            if c_stacks(k_id, k_stacks) >= 0 then
                for i = 0, 127 do
                    local addr = k_stacks.addr[i]
                    if addr ~= 0 then
                        k_arr[i + 1] = tostring(addr)
                    else
                        break
                    end
                end
            end
            reverseTable(k_arr)
        elseif k_id == -12 then
            overFlow = true
        end
        cell[3] = tonumber(k_id)
        cell[4] = k_arr

        local u_id = key.u_stack   -- process u_stack
        local u_arr = {}
        if u_id > 0 then
            local u_stacks = c_new("struct stack_addrs")
            if c_stacks(u_id, u_stacks) >= 0 then
                for i = 0, 127 do
                    local addr = u_stacks.addr[i]
                    if addr ~= 0 then
                        u_arr[i + 1] = tostring(addr)
                    else
                        break
                    end
                end
            end
            reverseTable(u_arr)
        elseif u_id == -12 then
            overFlow = true
        end
        cell[5] = tonumber(u_id)
        cell[6] = u_arr
        cell[7] = tonumber(v)

        cells[cnt] = cell
        cnt = cnt + 1
        ::continue::
    end
    self._cells[self._prefix] = cells
    return overFlow
end

return CbackTrace
