local mt = {}

local pystring = require("pystring")
local pstat = require("posix.sys.stat")
local fcntl = require("posix.fcntl")
local posix = require("posix")
local unistd = require("posix.unistd")
local system = require("common.system")

local assert = assert
local unpack = unpack
local error = error
local tostring = tostring
local execute = os.execute
local rsplit = pystring.rsplit
local pyfind = pystring.find
local stat = pstat.stat
local open = fcntl.open
local close = posix.close
local getpid = unistd.getpid
local pwrite = unistd.write
local pfcntl = fcntl.fcntl
local format = string.format
local liteAssert = system.liteAssert

function mt.lock(path)
    liteAssert(pyfind(path, "/"), format("path %s must be absolute", path))
    local dir, _ = unpack(rsplit(path, "/", 1))
    local res, err, errno = stat(dir)
    if not res then
        if errno == 2 then
            local cmd = format("mkdir -p %s", dir)
            liteAssert(execute(cmd), format("mkdir %s failed", dir))
        else
            error(err)
        end
    end
    res, err, errno = stat(dir)
    liteAssert(pstat.S_ISDIR(res.st_mode), format("%s is not a directory", dir))
    local fd = assert(open(path, fcntl.O_CREAT + fcntl.O_RDWR, 511))
    local lock = {
        l_type = posix.F_WRLCK;
        l_whence = posix.SEEK_SET;
        l_start = 0;
        l_len = 0; -- full lock
    }

    res, err, errno = pfcntl(fd, fcntl.F_SETLK, lock)
    if not res then
        close(fd)
        error(err .. "\n !!!one instance only!!!")
    end

    pwrite(fd, tostring(getpid()))

    return fd
end

function mt.unlock(fd)
    local unlock = { l_type = posix.F_UNLCK; l_whence = posix.SEEK_SET; l_start = 0; l_len = 0; }
    liteAssert(pfcntl(fd, fcntl.F_SETLK, unlock))
    close(fd)
end

return mt