local require = require

local ffi = require("ffi")
local system = require("common.system")
local workVar = require("module.workVar")
local pystring = require("pystring")
local socket = require("socket")
local gc = require("haWork.gc")

local tonumber = tonumber
local floor = math.floor
local format = string.format
local create = coroutine.create
local resume = coroutine.resume
local status = coroutine.status
local print = print
local math_huge = math.huge
local msleep = workVar.msleep
local coReport = system.coReport
local with = pystring.with
local gettime = socket.gettime
local gcfree = gc.free

ffi.cdef [[
    struct on_stacks{
        int k_stack;
        int u_stack;
        int pid;
    };
    
    struct stack_addrs{
        unsigned long addr[127];
    };
]]

local liveCos = {}

local mt = {}

local rundMaps = {}
local rundWorkList = {}

local function kernelVersion()
    local info = with("/proc/version")

    local major, minor = info:match("Linux version (%d+)%.(%d+)")
    return tonumber(major), tonumber(minor)
end
local major, minor = kernelVersion()

local period = 10000
local function next10sencond()
    local current = gettime() * 1000
    local next10 = current + 1 + period - (current % period)  -- 10s, to make sure next 10s, need add 1.
    return floor(next10 - current)
end

local function call(mod, args, module, loop)
    local func
    local o = mod.new(args, rundMaps)
    if major > 5 or (major == 5 and minor >= 6) then
        func = function() o:pre(); return o:calls() end
    else
        func = function() o:pre(); return o:call() end
    end

    local ret
    local next
    loop = loop > 0 and loop or math_huge
    repeat
        next = next10sencond()
        ret = msleep(next)
        if not ret or ret ~= next then  -- break from host.
            break
        end

        ret = func()
        if ret < 0 then
            break
        end
        loop = loop - 1
    until loop <= 0
    o:stop()
    print(format("thread %s stoped", module))
    liveCos[module] = nil
    rundWorkList[module] = false
    gcfree()
end

local function startThread(mod, args, module, loop)
    local co
    co = create(call)
    local ok, msg = resume(co, mod, args, module, loop)
    coReport(co, ok, msg)
    return co
end

function mt.start(module, args)
    if not module then
        return {body = "module is none."}
    end
    if liveCos[module] then
        return {body = format("liveTrace moudle %s is running, should stop at first.", module)}
    end

    local path = format("liveTrace.module.%s", module)
    local mod = require(path)
    if not mod then
        return {body = format("module %s is not installed.", module)}
    end
      
    local loop = args.loop or 100
    local lock = format("/tmp/sysom/lock/%s.lock", module)
    args.lock = lock
    local co = startThread(mod, args, module, loop)
    liveCos[module] = co
    rundWorkList[module] = true
    return {body = format("thread %s started.", module)}
end

function mt.stop(module)
    if not module then
        return {body = "module is none."}
    end
    if not liveCos[module] then
        return {body = format("liveTrace moudle %s is not running.", module)}
    end
    local co = liveCos[module]
    local ok, msg = resume(co)
    coReport(co, ok, msg)
    return {body = format("thread %s stopped.", module)}
end

function mt.getMaps(key)
    return rundMaps[key]
end

function mt.getWorList()
    return rundWorkList
end

return mt
