local require = require
require("eclass")
local flock = require("liveTrace.lib.flock")

local oncpuFfi = require("liveTrace.modffi.oncpuFfi")
local system = require("common.system")
local ffi, cffi = oncpuFfi.ffi, oncpuFfi.cffi
local CbackTrace = require("liveTrace.lib.backTrace")
local gc = require("haWork.gc")

local pairs = pairs
local c_init = cffi.on_init
local c_deinit = cffi.on_deinit
local c_stacks = cffi.on_stacks
local c_clear_hash = cffi.on_clear_hash
local c_fire_pre = cffi.on_fire_pre
local c_calls = cffi.on_calls
local c_call = cffi.on_call
local c_clear_stack = cffi.on_clear_stack
local c_new = ffi.new
local c_string = ffi.string
local c_sizeof = ffi.sizeof
local c_copy = ffi.copy
local liteAssert = system.liteAssert
local print = print
local gfree = gc.free

local class = class
local Concpu = class("oncpu")

function Concpu:_init_(args, elfdump)
    self._fdLock = flock.lock(args.lock)
    print(c_init())
    self._bpf = true
    args.storePrefix = "on"
    self._back = CbackTrace.new(elfdump, c_stacks, args)
end

function Concpu:_del_()
    self:stop()
end

function Concpu:stop()
    if self._fdLock then
        flock.unlock(self._fdLock)
        self._fdLock = nil
    end
    if self._bpf then
        c_deinit()
        self._bpf = false
    end
    self._back:stop()
end

function Concpu:pre()
    liteAssert(c_fire_pre() == 0, "oncpu pre failed.")
end

local function post()
    gfree()
end

local function saveStack(stacks, key, v)
    local key_copy = c_new("struct on_stacks")
    c_copy(key_copy, key, c_sizeof("struct on_stacks"))
    stacks[key_copy] = v
end

function Concpu:calls()
    local batch = 1024
    local cnt = 0
    local keys   = c_new("struct on_stacks[?]", batch)
    local values = c_new("int[?]", batch)
    local cntSyms = 0

    local res = 0
    local in_batch = c_new("void*[1]", ffi.NULL)
    local out_batch = c_new("void*[1]")
    local batch_read = c_new("unsigned int[1]", batch)
    local stacks = {}

    repeat
        res = c_calls(in_batch, out_batch, keys, values, batch_read)
        local bread = batch_read[0]
        cntSyms = cntSyms + bread
        for i = 0, bread - 1 do
            local key = keys[i]
            local v = values[i]
            if v > 0 then
                cnt = cnt + v
                saveStack(stacks, key, v)
            end
        end
        in_batch = out_batch
    until res~= 0

    c_clear_hash(keys[0])
    if self._back:back(stacks, true) then
        c_clear_stack()
    end
    post()
    return 0
end

function Concpu:call()
    local v = 0
    local cnt = 0
    local size = c_sizeof("struct on_stacks")
    local key  = c_new("struct on_stacks")
    local next = c_new("struct on_stacks")
    local cntSyms = 0

    local stacks = {}

    while true do
        v = c_call(key, next)
        if v > 0 then
            cnt = cnt + v
            cntSyms = cntSyms + 1
            saveStack(stacks, next, v)
            c_copy(key, next, size)
        else
            break
        end
    end

    for k, _ in pairs(stacks) do
        c_clear_hash(k)
    end
    if self._back:back(stacks, true) then
        c_clear_stack()
    end
    post()
    return 0
end

return Concpu